"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.patchTSModule = void 0;
var fs_1 = __importDefault(require("fs"));
var path_1 = __importDefault(require("path"));
var system_1 = require("./system");
var file_utils_1 = require("./file-utils");
var actions_1 = require("./actions");
/* ****************************************************************************************************************** */
// region: Constants
/* ****************************************************************************************************************** */
var dtsPatchSrc = '\n' +
    fs_1.default.readFileSync(path_1.default.resolve(system_1.appRoot, system_1.tspPackageJSON.directories.resources, 'module-patch.d.ts'), 'utf-8');
var jsPatchSrc = fs_1.default.readFileSync(path_1.default.resolve(system_1.appRoot, system_1.tspPackageJSON.directories.resources, 'module-patch.js'), 'utf-8');
// endregion
/* ********************************************************************************************************************
 * Helpers
 * ********************************************************************************************************************/
var getHeader = function () { return "/// tsp: ".concat(system_1.tspPackageJSON.version, "\n\n"); };
/**
 * Generate insertion code for module-patch
 */
var generatePatch = function (isTSC) {
    return jsPatchSrc +
        "\nObject.assign(tsp, { isTSC: ".concat(isTSC, ", tspVersion: '").concat(system_1.tspPackageJSON.version, "' });\n\n");
};
/**
 * Validate TSModule and TSPackage before patching
 */
function validate(tsModule, tsPackage) {
    if (tsModule) {
        var file = tsModule.file, filename = tsModule.filename, dir = tsModule.dir, patchVersion = tsModule.patchVersion, canPatch = tsModule.canPatch, outOfDate = tsModule.outOfDate;
        if (!fs_1.default.existsSync(file))
            throw new system_1.FileNotFound("Could not find module ".concat(filename, " in ").concat(dir + path_1.default.sep));
        if (patchVersion && !outOfDate)
            throw new system_1.PatchError("Module ".concat(filename, " is already up-to-date with local version - v").concat(patchVersion));
        if (!canPatch)
            throw new system_1.PatchError("Module ".concat(filename, " cannot be patched! No instance of TypeScript found."));
    }
    if (tsPackage) {
        var major = tsPackage.version.split('.')[0];
        if (+major < 4)
            throw new system_1.WrongTSVersion("ts-patch v2 requires TypeScript v4.0 or higher.");
    }
    return true;
}
var patchModuleDiagnostics = function (tsModule, tsPackage, source) {
    var src = source || tsModule.moduleSrc;
    var funcPos = src.search(/function emitFilesAndReportErrors\(/);
    if (funcPos < 0)
        throw new Error("Bad TS Code. Could not find function emitFilesAndReportErrors in ".concat(tsModule.filename));
    var startCode = src.substr(0, funcPos);
    var restCode = src.substr(funcPos);
    /* Find emit call position  */
    var emitPos = restCode.search(/^\s*?var emitResult =/m);
    if (emitPos < 0)
        throw new Error("Could not determine emit call. Please file an issue with your TS version!");
    var ver = tsPackage.version.split('.');
    var majorVer = +ver[0];
    var minorVer = +ver[1];
    var diagnosticsArrName = 'allDiagnostics';
    return startCode +
        restCode.substr(0, emitPos) +
        "\nts.diagnosticMap.set(program, ".concat(diagnosticsArrName, ");\n") +
        restCode.substr(emitPos);
};
/* ********************************************************************************************************************
 * Patch
 * ********************************************************************************************************************/
/**
 * Patch TypeScript Module
 */
function patchTSModule(tsModule, tsPackage) {
    validate(tsModule, tsPackage);
    var filename = tsModule.filename, file = tsModule.file, dir = tsModule.dir;
    /* Install patch */
    var isTSC = (filename === 'tsc.js');
    var patchSrc = generatePatch(isTSC);
    /* Add diagnostic modification support */
    var moduleSrc = patchModuleDiagnostics(tsModule, tsPackage);
    try {
        if (isTSC) {
            /* Select non-patched typescript.js */
            var tsFile = [
                path_1.default.join(tsPackage.packageDir, actions_1.BACKUP_DIRNAME, 'typescript.js'),
                path_1.default.join(tsPackage.libDir, 'typescript.js')
            ]
                .filter(function (f) { return fs_1.default.existsSync(f); })[0];
            /* Get TSC-specific module slice */
            var ver = tsPackage.version.split('.');
            var majorVer = +ver[0];
            var minorVer = +ver[1];
            var tscSlice = majorVer >= 4 && minorVer >= 9
                ? moduleSrc.replace(/^[\s\S]+(\(function \(ts\) {\s+var StatisticType;[\s\S]+)$/, '$1')
                : moduleSrc.replace(/^[\s\S]+(\(function \(ts\) {\s+function countLines[\s\S]+)$/, '$1');
            var execCmdPost = tscSlice.lastIndexOf('\nts.executeCommandLine\(');
            if (execCmdPost < 0)
                throw new Error("Could not find tsc executeCommandLine");
            var execCmd = tscSlice.slice(execCmdPost);
            tscSlice = tscSlice.slice(0, execCmdPost);
            /* Expand TSC with full typescript library (splice tsc part on top of typescript.ts code) */
            var content = Buffer.concat([
                Buffer.from(getHeader()),
                Buffer.from(patchModuleDiagnostics(tsModule, tsPackage, fs_1.default.readFileSync(tsFile, 'utf-8'))),
                Buffer.from(!(0, file_utils_1.getTSModule)(tsFile).patchVersion ? patchSrc : ''),
                Buffer.from(tscSlice),
                Buffer.from(execCmd)
            ]);
            fs_1.default.writeFileSync(file, content);
        }
        else
            fs_1.default.writeFileSync(file, Buffer.concat([
                Buffer.from(getHeader()),
                Buffer.from(moduleSrc),
                Buffer.from(patchSrc)
            ]));
    }
    catch (e) {
        throw new system_1.FileWriteError(filename, e.message);
    }
    /* Patch d.ts with types (if module is typescript.ts) */
    if (filename === 'typescript.js') {
        var targetFile = path_1.default.join(dir, 'typescript.d.ts');
        var backupFile = path_1.default.join(tsPackage.packageDir, actions_1.BACKUP_DIRNAME, 'typescript.d.ts');
        try {
            if (fs_1.default.existsSync(backupFile))
                fs_1.default.writeFileSync(targetFile, fs_1.default.readFileSync(backupFile, 'utf-8') + dtsPatchSrc);
            else
                fs_1.default.appendFileSync(targetFile, dtsPatchSrc);
        }
        catch (e) {
            throw new system_1.FileWriteError(filename, e.stack);
        }
    }
}
exports.patchTSModule = patchTSModule;
//# sourceMappingURL=patcher.js.map